package de.hftstuttgart.unifiedticketing.systems.gitlab;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.hftstuttgart.unifiedticketing.core.Filter;
import de.hftstuttgart.unifiedticketing.core.Logging;
import de.hftstuttgart.unifiedticketing.core.TicketSystem;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.DeserializationException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpReqeustException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpResponseException;
import okhttp3.*;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class GitlabFilter extends Filter<GitlabTicket, GitlabFilter>
{
    private static Logger logger = Logging.getLogger(GitlabFilter.class.getName());
    protected final GitlabTicketSystem parent;

    protected GitlabFilter(GitlabTicketSystem parent)
    {
        this.parent = parent;
    }

    protected OkHttpClient getHttpClient() { return new OkHttpClient(); }

    /**
     * this builds up a request with as much of the given constraints directed to the
     * gitlab api. Everything not possible through the api will be filtered afterwards in java
     * @return ticket list matching given constraints
     */
    @Override
    public List<GitlabTicket> get()
    {
        OkHttpClient client;
        ObjectMapper mapper = new ObjectMapper()
            .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        Request.Builder requestBuilder = new Request.Builder()
            .url(String.format("%s/issues", parent.baseUrl))
            .get();

        if (parent.apiKey != null)
        {
            requestBuilder.addHeader("PRIVATE-TOKEN", parent.apiKey);
        }

        HttpUrl.Builder urlBuilder = requestBuilder.build().url().newBuilder();

        for (Map.Entry<String, Object> mapEntry : setFilters.entrySet())
        {
            String f = mapEntry.getKey();
            Object v = mapEntry.getValue();
            try
            {
                if (f.equals(FilterNames.ASSIGNEEID.name()))
                {
                    ((Set<String>) v).stream()
                        .findFirst()
                        .ifPresent(value -> urlBuilder.addQueryParameter("assignee_id", value));

                } else if (f.equals(FilterNames.ASSIGNEENAME.name()))
                {
                    urlBuilder.addQueryParameter("assignee_username", ((Set<String>) v).stream()
                        .reduce((u1, u2) -> u1 + "," + u2)
                        .orElse(""));
                } else if (f.equals(FilterNames.DESCRIPTION_CONTAIN.name()))
                {
                    urlBuilder.addQueryParameter("search", (String) v);
                } else if (f.equals(FilterNames.DESCRIPTION_MATCH.name()))
                {
                    logger.log(Level.FINE, "Regex matching only possible after request |" +
                        " Filter: " + FilterNames.DESCRIPTION_MATCH.name());
                } else if (f.equals(FilterNames.IDS.name()))
                {
                    urlBuilder.addQueryParameter("iids", ((Set<String>) v).stream()
                        .reduce((id1, id2) -> id1 + "," + id2)
                        .orElse(""));
                } else if (f.equals(FilterNames.LABELS.name()))
                {
                    urlBuilder.addQueryParameter("labels", ((Set<String>) v).stream()
                        .reduce((l1, l2) -> l1 + "," + l2)
                        .orElse(""));
                } else if (f.equals(FilterNames.PAGE.name()))
                {
                    urlBuilder.addQueryParameter("page", String.valueOf(v));
                } else if (f.equals(FilterNames.PAGINATION.name()))
                {
                    urlBuilder.addQueryParameter("per_page", String.valueOf(v));
                } else if (f.equals(FilterNames.OPEN.name()))
                {
                    urlBuilder.addQueryParameter("state", ((boolean) v) ? "opened" : "closed");
                } else if (f.equals(FilterNames.TITLE_CONTAINS.name()))
                {
                    urlBuilder.addQueryParameter("search", (String) v);
                } else if (f.equals(FilterNames.TITLE_MATCH.name()))
                {
                    logger.log(Level.FINE, "Regex matching only possible after request |" +
                        " Filter: " + FilterNames.TITLE_MATCH.name());
                } else
                {
                    logger.log(Level.WARNING, String.format("unrecognized filter key: %s", f));
                }
            }
            catch (ClassCastException e)
            {
                logger.log(Level.SEVERE, "Filter with key "
                    + f
                    + " unexpectedly had type "
                    + v.getClass().getName());
                if (!this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR))
                {
                    throw new AssertionException(e);
                }
            }
        }

        requestBuilder.url(urlBuilder.build());

        client = getHttpClient();
        Response response;
        try
        {
            Request request = requestBuilder.build();
            logger.log(Level.FINEST, String.format(
                "created request:\n%s",
                (this.parent.apiKey != null)
                    ? request.toString().replace(this.parent.apiKey, "SECRET")
                        : request.toString()
            ));

            response = client.newCall(request).execute();
        }
        catch (IOException e)
        {
            logger.log(Level.SEVERE, String.format("get request FAILED with: %s", e.getMessage()));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpReqeustException(e);
        }

        if (response.code() >= 400)
        {
            logger.log(Level.SEVERE, String.format(
                "request failed with response code %d",
                response.code()
            ));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpResponseException(
                String.format("ticket query failed, error response code: %d", response.code()),
                response.code());
        }

        logger.log(Level.FINEST, "response received\n");

        ResponseBody responseBody;
        responseBody = response.body();

        if (responseBody == null)
        {
            logger.log(Level.SEVERE, "query didn't deliver a body in response");

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new HttpResponseException("ticket query failed, no response body", response.code());
        }

        List<GitlabTicketResponse> tr;
        try
        {
            tr = mapper.readValue(responseBody.bytes(), new TypeReference<List<GitlabTicketResponse>>(){});

            logger.log(Level.FINER, "parsed response body to ticketResponse list instance");
            logger.log(Level.FINEST, String.format("found %d items pre post-filter", tr.size()));
        } catch (IOException e)
        {
            logger.log(Level.SEVERE, String.format("parsing query response FAILED with: %s", e.getMessage()));

            if (this.parent.getConfigTrue(TicketSystem.ConfigurationOptions.RETURN_NULL_ON_ERROR)) return null;
            else throw new DeserializationException(e);
        }

        lastReceivedItemCount = tr.size();

        logger.log(Level.FINER, "starting query post filter");
        Stream<GitlabTicketResponse> ticketStream = tr.stream();

        for (Map.Entry<String, Object> entry : setFilters.entrySet())
        {
            String f = entry.getKey();
            Object v = entry.getValue();
            try
            {
                if (f.equals(FilterNames.DESCRIPTION_MATCH.name()))
                {
                    ticketStream = ticketStream.filter(t -> t.description.matches((String) v));
                }
                else if (f.equals(FilterNames.TITLE_MATCH.name()))
                {
                    ticketStream = ticketStream.filter(t -> t.title.matches((String) v));
                }
            } catch (ClassCastException e)
            {
                logger.log(Level.SEVERE, "Filter with key "
                    + f
                    + " unexpectedly had type "
                    + v.getClass().getName());
            }
        }

        logger.log(Level.FINER, "post-filter finished");

        List<GitlabTicket> ret = ticketStream.map(t -> GitlabTicket.fromTicketResponse(parent, t))
            .collect(Collectors.toList());
        logger.log(Level.FINEST, String.format("remaining items: %d", ret.size()));

        return ret;
    }
}
